{% trans %}
<p>
    Anonymous usage tracking provides valuable insights into how many instances running what versions of OctoPrint are out there,
    whether they are successfully completing print jobs and various other metrics.
</p>
<p>
    By enabling it you help to identify problems with new releases and release candidates early on, and to better
    tailor OctoPrint's future development to actual use.
</p>
{% endtrans %}
