<form class="form-horizontal" onsubmit="return false;">
    <div class="control-group" title="{{ _('Maximum size for which the GCode Viewer autoloads the file for preview')|edq }}">
        <label class="control-label" for="settings-featureSizeThreshold">{{ _('GCode Viewer file size threshold') }}</label>
        <div class="controls">
            <div class="input-append">
                <input type="text" class="input-mini text-right" data-bind="value: sizeThresholdStr" id="settings-featureSizeThreshold">
            </div>
            <span class="help-inline">{{ _('on desktop')}}</span>
        </div>
        <div class="controls">
            <div class="input-append">
                 <input type="text" class="input-mini text-right" data-bind="value: mobileSizeThresholdStr">
            </div>
            <span class="help-inline">{{ _('on mobile')}}</span>
        </div>
        <div class="controls">
            <span class="help-inline">
                {% trans %}Provide values including size unit. Allowed units are: b, byte, bytes, kb, mb, gb, tb (case insensitive). Example: <code>5MB</code>, <code>500KB</code>{% endtrans %}
            </span>
        </div>
    </div>
</form>
