<div class="control-group">
    <div class="controls">
        <label class="checkbox">
            <input type="checkbox" data-bind="checked: webcam_flipH" id="settings-webcamFlipH"> {{ _('Flip webcam horizontally') }}
        </label>
    </div>
    <div class="controls">
        <label class="checkbox">
            <input type="checkbox" data-bind="checked: webcam_flipV" id="settings-webcamFlipV"> {{ _('Flip webcam vertically') }}
        </label>
    </div>
    <div class="controls">
        <label class="checkbox">
            <input type="checkbox" data-bind="checked: webcam_rotate90" id="settings-webcamRotate90"> {{ _('Rotate webcam 90 degrees counter clockwise') }}
        </label>
    </div>
</div>
