<h3>{{ _('Restore from a backup?') }}</h3>

{% trans %}<p>
    If you would like to restore OctoPrint configuration from a backup now is the time.
</p>{% endtrans %}

{% set plugin_backup_upload_form_id='wizard-backup-upload' %}
{% include "snippets/plugin/backup/backup_plugin_upload_form.jinja2" %}
