<div class="control-group" title="{{ _('Bitrate to use for encoding the timelapse video')|edq }}">
    <label class="control-label" for="settings-webcamBitrate">{{ _('Timelapse bitrate') }}</label>
    <div class="controls">
        <div class="input-append">
            <input type="text" class="input-mini text-right" data-bind="value: webcam_bitrate" id="settings-webcamBitrate">
            <span class="add-on">bit/s</span>
        </div>
    </div>
</div>
